/*
 * This is a test program for Rexx/SQL.  It demonstrates how to handle
 * "BLOB" datatypes using the ODBC interface.
 *
 * It creates a table with various "normal" columns, with a "long" varchar
 * column and a "long" binary column.  The file pointed to by the 
 * text_filename parameter is inserted into the "long" varchar column, 
 * while the file pointed to by the binary_filename parameter is inserted
 * into the "long" binary column.
 *
 * After inserting a row with this data, the table is queried and new
 * files are created with a ".out" appended to the input file names.
 * The new files should be identical to the input files.
 *
 * Run this as:
 * rexxsql blob.cmd user=username pass=password data=database text=text_filename binary=binary_filename
 *
 * user= specifies the username to log into the database
 * user= specifies the password to log into the database
 * data= specifies the database DSN to connect to
 * text= specifies a text file which will be loaded into the database
 * binary= specifies a binary file which will be loaded into the database
 *
 * Two variables; long_binary_type and long_text_type define the datatypes
 * of the long binary and long text columns. The database "integer" type
 * is also easily changeable. Change these if necessary.
 *
 * This program will unload the Rexx/SQL external functions on exit,
 * so don't run this if unloading the external functions could cause
 * other running Rexx/SQL programs to fail.
 */
/* 
long_binary_type = 'long binary'
long_text_type = 'long varchar'
int_type = 'integer'
*/
long_binary_type = 'image'
long_text_type = 'text'
int_type = 'int'

Signal On Syntax
Parse Arg . 'user=' username . 1 . 'pass=' password . 1 . 'data=' database . 1 . 'text=' text_filename . 1 . 'binary=' binary_filename .
If username = '' & password = '' & database = '' & text_filename = '' & binary_filename = '' Then
  Do
     Say 'Usage: blob.cmd user=username pass=password data=database text=text_filename binary=binary_filename'
     Exit 1
  End
If RxFuncQuery( 'SQLLoadFuncs' ) Then
  Do
     Call RxFuncAdd 'SQLLoadFuncs', 'rexxsql', 'SQLLoadFuncs'
     Call SQLLoadFuncs
  End
text_size = Stream( text_filename, 'C', 'QUERY SIZE' )
If text_size = '' Then Abort( text_filename 'does not exist' )
binary_size = Stream( binary_filename, 'C', 'QUERY SIZE' )
If binary_size = '' Then Abort( binary_filename 'does not exist' )
blocksize = SQLVariable( 'LONGLIMIT' )
/*
 * Connect and display some details of the connection
 */
If SQLConnect( 'c1', username, password, database ) < 0 Then Abort( 'connecting' )
Say 'Connect succeeded!'
Say 'Rexx/SQL Version:' SQLVariable( 'VERSION' )
If SQLGetinfo( 'c1', 'DBMSNAME', 'desc.' ) < 0 Then Abort( 'getting db version' )
Say 'Database Name:   ' desc.1
If SQLGetinfo( 'c1', 'DBMSVERSION', 'desc.') < 0 Then Abort( 'getting db version' )
Say 'Database Version:' desc.1
/*
 * Create the test table
 */
create = 'create table rexxsql_testblob (filesize_text' int_type ', filesize_binary' int_type ', filename_text varchar(50), filename_binary varchar(50), file_text' long_text_type ', file_binary' long_binary_type ')'
Call SQLCommand 'd1', 'drop table rexxsql_testblob'
If SQLCommand( 'c1', create ) < 0 Then Abort( 'creating rexxsql_testblob table' )
/*
 * Insert the data
 */
If SQLPrepare( 'i1', 'insert into rexxsql_testblob values ( ?, ?, ?, ?, ?, ? )' ) < 0 Then Abort( 'preparing INSERT statement' )
If SQLExecute( 'i1', int_type, text_size, int_type, binary_size, 'CHAR', text_filename, 'CHAR', binary_filename, 'LONG:' || long_text_type, text_filename, 'LONG:' || long_binary_type , binary_filename ) < 0 Then Abort( 'executing INSERT statement' )
If SQLClose( 'i1' ) < 0 Then Abort( 'closing INSERT statement' )
/*
 * Extract the data
 */
If SQLPrepare( 'q1', 'select * from rexxsql_testblob', 'nobind') < 0 Then Abort( 'preparing SELECT statement' )
If SQLOpen( 'q1' ) < 0 Then Abort( 'opening SELECT statement' )
num = 0
Do Forever
   rc = SQLFetch( 'q1' )
   If rc = 0 Then Leave
   /*
    * We now have to get the column data for each column in the select
    * statement in the order they are returned from the database. This is
    * a limitation of most ODBC drivers!
    */
   If SQLGetData( 'q1', 'filesize_text' ) < 0 Then Abort( 'getting data for filesize_text column' )
   If SQLGetData( 'q1', 'filesize_binary' ) < 0 Then Abort( 'getting data for filesize_binary column' )
   If SQLGetData( 'q1', 'filename_text' ) < 0 Then Abort( 'getting data for filename_text column' )
   If SQLGetData( 'q1', 'filename_binary' ) < 0 Then Abort( 'getting data for filename_binary column' )
   num = num + 1
   Call Stream q1.filename_text || '.out', 'C', 'OPEN WRITE REPLACE'
   /*
    * Get the contents of the long text column and write to a file
    */
   Do i = 1 
      rc = SQLGetdata( 'q1', 'file_text', 1+(blocksize*(i-1)), blocksize )
      if rc < 0 Then Abort( 'getting data for file_text column; from' i 'for' blocksize )
      If rc == 0 Then Leave
      Say 'Got' rc 'bytes for' q1.filename_text || '.out'
      Call Charout q1.filename_text || '.out', q1.file_text
   End
   Call Stream q1.filename_text || '.out', 'C', 'CLOSE'
   Call Stream q1.filename_binary || '.out', 'C', 'OPEN WRITE REPLACE'
   /*
    * Get the contents of the long binary column
    */
   Do i = 1 
      rc = SQLGetdata( 'q1', 'file_binary', 1+(blocksize*(i-1)), blocksize )
      if rc < 0 Then Abort( 'getting data for file_binary column; from' i 'for' blocksize )
      If rc == 0 Then Leave
      Say 'Got' rc 'bytes for' q1.filename_binary || '.out'
      Call Charout q1.filename_binary || '.out', q1.file_binary
   End
   Call Stream q1.filename_binary || '.out', 'C', 'CLOSE'
   Say 'Retreived row: filesize_text:' q1.filesize_text 'filesize_binary:' q1.filesize_binary 'filename_text:' q1.filename_text 'filename_binary:' q1.filename_binary
End
If SQLClose( 'q1' ) < 0 Then Abort( 'closing SELECT statement' )
/*
 * All done, lets get out
 */
If SQLDisconnect( 'c1' ) < 0 Then Abort( 'disconnecting' )
Say 'Disconnect succeeded!'
Call SQLDropFuncs 'UNLOAD'
Return

Abort: Procedure Expose sqlca.
Parse Arg msg
Say 'Program failed:' msg
Say sqlca.interrm
If Datatype( sqlca.sqlerrm.0 ) = 'NUM' Then
   Do i = 1 To sqlca.sqlerrm.0
      Say '('sqlca.sqlstate.i')' sqlca.sqlerrm.i
   End
Else
   Do
      Say '('sqlca.sqlstate')' sqlca.sqlerrm
   End
Call SQLDropFuncs 'UNLOAD'
Exit 1

Syntax: 
Call Abort 'syntax error on line' sigl
Return
